//----------------------------------------------------------------------------
//
// File:        opcodes.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <stdio.h>
#include "common.hpp"

USHORT  *PC;
USHORT   curOpCode;

ULONG    srcComp;
ULONG    dstComp;
UCHAR    flagLogical;
UCHAR    flagArithmetic;
UCHAR    flagEqual;
UCHAR    flagCarry;
UCHAR    flagOverflow;
UCHAR    flagParity;
UCHAR    flagXOP;
UCHAR    intLevel;


UCHAR    parity [ 256 ];

static sOpCode *op;

void opcode_A ()
{
    USHORT srcAddress = GetAddress ( curOpCode );
    USHORT src = ReadMemoryW ( srcAddress );
    USHORT dstAddress = GetAddress ( curOpCode >> 6 );
    ULONG  dst = ReadMemoryW ( destAddress );

    dst += src;

    flagCarry    = ( dst & 0x00080000 ) ? 1 : 0;
    flagOverflow = ??:
    srcComp      = dst;
    dstComp      = 0;

    WriteMemoryW ( destAddress, dst );
}

void opcode_AB ()
{
    USHORT dstAddress = GetAddress ( curOpCode >> 6 );
    USHORT dst = ReadMemoryB ( dstAddress );
    USHORT srcAddress = GetAddress ( curOpCode );
    USHORT src = ReadMemoryB ( srcAddress );

    dst += src;

    flagCarry    = ( dst & 0x00080000 ) ? 1 : 0;
    flagOverflow = ??;
    flagParity   = parity [ ( UCHAR ) dst ];
    srcComp      = dst;
    dstComp      = 0;

    WriteMemoryB ( destAddress, dst );
}

void opcode_ABS ()
{
    USHORT dstAddress = GetAddress ( curOpCode );
    USHORT dst = ReadMemoryW ( address );

    dstComp      = dst;
    dstComp      = 0;

    if ( dst & 0x8000 ) {
        flagCarry    = ( dst & 0x00080000 ) ? 1 : 0;
        WriteMemoryW ( dstAddress, dst );
    } else {
        flagOverflow = 0;
    }
}

void opcode_AI ()
{
    USHORT reg = curOpCode & 0x000F;
    USHORT value = ReadRegisterW ( reg ) + *PC++;

    flagCarry    = ( value & 0x00080000 ) ? 1 : 0;
    srcComp      = value;
    dstComp      = 0;
}

void opcode_DEC ()
{
    USHORT address = GetAddress ( curOpCode );
    _AX = ReadMemoryW ( address );
l	add	ax,-1;
    USHORT value = _AX;
l	setc	cl;			// Carry
l	shl	cl,1;
l	seto	ch;			// Overflow
l	or	cl,ch;
l	shl	cx,11;
    ST.value &= 0xE7FF;
    ST.value |= _CX;
    WriteMemoryW ( address, value );
    CompareW ( value, 0 );
}

void opcode_DECT ()
{
    USHORT address = GetAddress ( curOpCode );
    _AX = ReadMemoryW ( address );
l	add	ax,-2;
    USHORT value = _AX;
l	setc	cl;			// Carry
l	shl	cl,1;
l	seto	ch;			// Overflow
l	or	cl,ch;
l	shl	cx,11;
    ST.value &= 0xE7FF;
    ST.value |= _CX;
    WriteMemoryW ( address, value );
    CompareW ( value, 0 );
}

void opcode_DIV ()
{
    _AX = GetAddress ( curOpCode );
    USHORT divisor = ReadMemoryW ( _AX );
    int reg = ( curOpCode >> 6 ) & 0x000F;
    USHORT value = ReadRegisterW ( reg );
    ST.flags.Overflow = value <= divisor;
    if ( ! ST.flags.Overflow ) {
        _AX = ReadRegisterW ( reg + 1 );
        _DX = value;
        _CX = divisor;
        asm div cx;
        value = _DX;
        WriteRegisterW ( reg, ( USHORT ) _AX );
        WriteRegisterW ( reg + 1, value );
    }
}

void opcode_INC ()
{
    USHORT address = GetAddress ( curOpCode );
    _AX = ReadMemoryW ( address );
l	add	ax,1;
    USHORT value = _AX;
l	setc	cl;			// Carry
l	shl	cl,1;
l	seto	ch;			// Overflow
l	or	cl,ch;
l	shl	cx,11;
    ST.value &= 0xE7FF;
    ST.value |= _CX;
    WriteMemoryW ( address, value );
    CompareW ( value, 0 );
}

void opcode_INCT ()
{
    USHORT address = GetAddress ( curOpCode );
    _AX = ReadMemoryW ( address );
l	add	ax,2;
    USHORT value = _AX;
l	setc	cl;			// Carry
l	shl	cl,1;
l	seto	ch;			// Overflow
l	or	cl,ch;
l	shl	cx,11;
    ST.value &= 0xE7FF;
    ST.value |= _CX;
    WriteMemoryW ( address, value );
    CompareW ( value, 0 );
}

void opcode_MPY ()
{
    _AX = GetAddress ( curOpCode );
    USHORT value = ReadMemoryW ( _AX );
    int reg = ( curOpCode >> 6 ) & 0x000F;
    _AX = ReadRegisterW ( reg );
    _DX = value;
    asm mul dx;
    value = _AX;
    WriteRegisterW ( reg, ( USHORT ) _DX );
    WriteRegisterW ( reg + 1, value );
}

void opcode_NEG ()
{
    USHORT address = GetAddress ( curOpCode );
    USHORT value = - ReadMemoryW ( address );
    ST.flags.Carry = value == 0;
    ST.flags.Overflow = value == 0x8000;
    WriteMemoryW ( address, value );
    CompareW ( value, 0 );
}

void opcode_S ()
{
    USHORT dst = GetAddress ( curOpCode >> 6 );
    _DI = ReadMemoryW ( dst );
    _AX = GetAddress ( curOpCode );
    _AX = ReadMemoryW ( _AX );
l	sub	di,ax;
l	setnc	al;			// Carry
l	shl	al,1;
l	seto	ah;			// Overflow
l	or	al,ah;
l	shl	ax,11;
    ST.value &= 0xE7FF;
    ST.value |= _AX;
    WriteMemoryW ( dst, ( USHORT ) _DI );
    CompareW (( USHORT ) _DI, 0 );
}

void opcode_SB ()
{
    USHORT dst = GetAddress ( curOpCode >> 6 );
    _DI = ReadMemoryB ( dst );
    _AX = GetAddress ( curOpCode );
    _AL = ReadMemoryB ( _AX );
l	mov	bx,di;
l	sub	bl,al;
l	mov	di,bx;
l	setnc	al;			// Carry
l	shl	al,1;
l	seto	ah;			// Overflow
l	or	al,ah;
l	shl	ax,11;
    ST.value &= 0xE7FF;
    ST.value |= _AX;
    WriteMemoryB ( dst, ( UCHAR ) _DI );
    CompareB (( UCHAR ) _DI, 0 );
}

// Jump & Branch Instructions

#define DISPLACEMENT(x)	((( char ) (( x ) & 0x00FF )) * 2 )

void opcode_B ()
{
    PC = GetAddress ( curOpCode );
}

void opcode_BL ()
{
    WriteRegisterW ( 11, PC );
    PC = GetAddress ( curOpCode );
}

void opcode_BLWP ()
{
    USHORT address = GetAddress ( curOpCode );
    USHORT newWP = _ReadMemoryW_ ( address );
    USHORT newPC = _ReadMemoryW_ ( address + 2 );
    ContextSwitch ( newWP, newPC );
}

void opcode_JEQ ()
{
    if ( ST.flags.Equal == 1 ) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JGT ()
{
    if ( ST.flags.ArithGt == 1 ) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JHE ()
{
    if (( ST.flags.LogicGt == 1 ) || ( ST.flags.Equal == 1 )) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JH ()
{
    if (( ST.flags.LogicGt == 1 ) && ( ST.flags.Equal == 0 )) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JL ()
{
    if (( ST.flags.LogicGt == 0 ) && ( ST.flags.Equal == 0 )) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JLE ()
{
    if (( ST.flags.LogicGt == 0 ) || ( ST.flags.Equal == 1 )) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JLT ()
{
    if (( ST.flags.ArithGt == 0 ) && ( ST.flags.Equal == 0 )) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JMP ()
{
    PC += DISPLACEMENT ( curOpCode );
}

void opcode_JNC ()
{
    if ( ST.flags.Carry == 0 ) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JNE ()
{
    if ( ST.flags.Equal == 0 ) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JNO ()
{
    if ( ST.flags.Overflow == 0 ) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JOP ()
{
    if ( ST.flags.OddParity == 1 ) PC += DISPLACEMENT ( curOpCode );
}

void opcode_JOC ()
{
    if ( ST.flags.Carry == 1 ) PC += DISPLACEMENT ( curOpCode );
}

void opcode_RTWP ()
{
    ST.value = ReadRegisterW ( 15 );
    PC = ReadRegisterW ( 14 );
    WP = ReadRegisterW ( 13 );
}

void opcode_X ()
{
    curOpCode = _ReadMemoryW_ ( GetAddress ( curOpCode ));
    ExecuteInstruction ();
    ((void(*)(void*)) op->function )( this );
}

void opcode_XOP ()
{
    USHORT address = GetAddress ( curOpCode );
    int level = (( curOpCode >> 4 ) & 0x003C ) + 64;
    USHORT newWP = _ReadMemoryW_ ( level );
    USHORT newPC = _ReadMemoryW_ ( level + 2 );
    ContextSwitch ( newWP, newPC );
    WriteRegisterW ( 11, address );
    ST.flags.XOP = true;
}

// Compare Instructions

void opcode_C ()
{
    USHORT src = GetAddress ( curOpCode );
    USHORT dst = GetAddress ( curOpCode >> 6 );
    CompareW ( ReadMemoryW ( src ), ReadMemoryW ( dst ));
}

void opcode_CB ()
{
    USHORT src = GetAddress ( curOpCode );
    USHORT dst = GetAddress ( curOpCode >> 6 );
    CompareB ( ReadMemoryB ( src ), ReadMemoryB ( dst ));
}

void opcode_CI ()
{
    CompareW ( ReadRegisterW ( curOpCode & 0x000F ), arg1 );
}

void opcode_COC ()
{
    _AX = GetAddress ( curOpCode );
    USHORT value1 = ReadMemoryW ( _AX );
    USHORT value2 = ReadRegisterW (( curOpCode >> 6 ) & 0x000F );
    ST.flags.Equal = ( value1 & value2 ) == value1;
}

void opcode_CZC ()
{
    _AX = GetAddress ( curOpCode );
    USHORT value1 = ReadMemoryW ( _AX );
    USHORT value2 = ~ ReadRegisterW (( curOpCode >> 6 ) & 0x000F );
    ST.flags.Equal = ( value1 & value2 ) == value1;
}

// Control & CRU Instructions

void opcode_LDCR ()
{
    USHORT address = GetAddress ( curOpCode ) & 0xFFFE;
    USHORT value = ReadMemoryW ( address );
    int cru = ReadRegisterW ( 12 ) >> 1;
    int count = ( curOpCode >> 6 ) & 0x000F;
    if ( count == 0 ) count = 16;
    if ( count < 9 ) {
        value >>= 8;
        CompareB (( UCHAR ) value, 0 );
    } else {
        CompareW ( value, 0 );
    }
    WriteCRU ( cru, count, value );
}

void opcode_SBO ()
{
    int cru = ( ReadRegisterW ( 12 ) >> 1 ) + ( curOpCode & 0x00FF );
    WriteCRU ( cru, 1, 1 );
}

void opcode_SBZ ()
{
    int cru = ( ReadRegisterW ( 12 ) >> 1 ) + ( curOpCode & 0x00FF );
    WriteCRU ( cru, 1, 0 );
}

void opcode_STCR ()
{
    USHORT address = GetAddress ( curOpCode ) & 0xFFFE;
    int cru = ReadRegisterW ( 12 ) >> 1;
    int count = ( curOpCode >> 6 ) & 0x000F;
    if ( count == 0 ) count = 16;
    if ( count < 9 ) {
        UCHAR value = ReadMemoryB ( address );
        value = ReadCRU ( cru, count );
        CompareB ( value, 0 );
        WriteMemoryB ( address, value );
    } else {
        USHORT value = ReadMemoryB ( address ) >> count;
        value = ReadCRU ( cru, count );
        CompareW ( value, 0 );
        WriteMemoryW ( address, value );
    }
}

void opcode_TB ()
{
    int cru = ( ReadRegisterW ( 12 ) >> 1 ) + ( curOpCode & 0x00FF );
    ST.flags.Equal = ReadCRU ( cru, 1 ) ? 1 : 0;
}

// Load & Move Instructions

void opcode_LI ()
{
    USHORT dst = *PC++;

    srcComp      = dst;
    dstComp      = 0;

    WriteRegisterW ( curOpCode & 0x000F, dst );
}

void opcode_LIMI ()
{
    ST.flags.IntMask = arg1;
}

void opcode_LWPI ()
{
    WP = arg1;
}

void opcode_MOV ()
{
    _AX = GetAddress ( curOpCode );
    USHORT value = ReadMemoryW ( _AX );
    _DI = GetAddress ( curOpCode >> 6 );
    WriteMemoryW ( _DI, value );
    CompareW ( value, 0 );
}

void opcode_MOVB ()
{
    _AX = GetAddress ( curOpCode );
    UCHAR value = ReadMemoryB ( _AX );
    _DI = GetAddress ( curOpCode >> 6 );
    WriteMemoryB ( _DI, value );
    CompareB ( value, 0 );
}

void opcode_STST ()
{
    PC -= 2;
    WriteRegisterW ( curOpCode & 0x000F, ST.value );
}

void opcode_STWP ()
{
    PC -= 2;
    WriteRegisterW ( curOpCode & 0x000F, WP );
}

void opcode_SWPB ()
{
    USHORT address = GetAddress ( curOpCode );
    _AX = ReadMemoryW ( address );
    asm xchg al,ah;
    WriteMemoryW ( address, ( USHORT ) _AX );
}

// Logical Instructions

void opcode_ANDI ()
{
    USHORT reg = curOpCode & 0x000F;
    USHORT value = ReadRegisterW ( reg ) & arg1;
    WriteRegisterW ( reg, value );
    CompareW ( value, 0 );
}

void opcode_ORI ()
{
    USHORT reg = curOpCode & 0x000F;
    USHORT value = ReadRegisterW ( reg ) | arg1;
    WriteRegisterW ( reg, value );
    CompareW ( value, 0 );
}

void opcode_XOR ()
{
    int reg = ( curOpCode >> 6 ) & 0x000F;
    USHORT src = GetAddress ( curOpCode );
    USHORT value = ReadRegisterW ( reg ) ^ ReadMemoryW ( src );
    WriteRegisterW ( reg, value );
    CompareW ( value, 0 );
}

void opcode_INV ()
{
    USHORT address = GetAddress ( curOpCode );
    USHORT value = ~ ReadMemoryW ( address );
    WriteMemoryW ( address, value );
    CompareW ( value, 0 );
}

void opcode_CLR ()
{
    _DI = GetAddress ( curOpCode );
    WriteMemoryW ( _DI, ( USHORT ) 0 );
}

void opcode_SETO ()
{
    _DI = GetAddress ( curOpCode );
    WriteMemoryW ( _DI, ( USHORT ) -1 );
}

void opcode_SOC ()
{
    _DI = GetAddress ( curOpCode >> 6 );
    _AX = GetAddress ( curOpCode );
    USHORT value = ReadMemoryW ( _AX ) | ReadMemoryW ( _DI );
    WriteMemoryW ( _DI, value );
    CompareW ( value, 0 );
}

void opcode_SOCB ()
{
    _DI = GetAddress ( curOpCode >> 6 );
    _AX = GetAddress ( curOpCode );
    UCHAR value = ReadMemoryB ( _AX ) | ReadMemoryB ( _DI );
    WriteMemoryB ( _DI, value );
    CompareB ( value, 0 );
}

void opcode_SZC ()
{
    _DI = GetAddress ( curOpCode >> 6 );
    _AX = GetAddress ( curOpCode );
    USHORT value = ~ReadMemoryW ( _AX ) & ReadMemoryW ( _DI );
    WriteMemoryW ( _DI, value );
    CompareW ( value, 0 );
}

void opcode_SZCB ()
{
    _DI = GetAddress ( curOpCode >> 6 );
    _AX = GetAddress ( curOpCode );
    UCHAR value = ~ReadMemoryB ( _AX ) & ReadMemoryB ( _DI );
    WriteMemoryB ( _DI, value );
    CompareB ( value, 0 );
}

// Workspace Register Shift Instructions

void opcode_SRA ()
{
    int reg = curOpCode & 0x000F;
    int count = ( curOpCode >> 4 ) & 0x000F;
    if ( count == 0 ) {
        count = ReadRegisterW ( 0 ) & 0x000F;
        if ( count == 0 ) count = 16;
    }
    short value = (( short ) ReadRegisterW ( reg )) >> --count;
    ST.flags.Carry = value;
    value >>= 1;
    WriteRegisterW ( reg, ( USHORT ) value );
    CompareW (( USHORT ) value, 0 );
}

void opcode_SRL ()
{
    int reg = curOpCode & 0x000F;
    int count = ( curOpCode >> 4 ) & 0x000F;
    if ( count == 0 ) {
        count = ReadRegisterW ( 0 ) & 0x000F;
        if ( count == 0 ) count = 16;
    }
    USHORT value = ReadRegisterW ( reg ) >> --count;
    ST.flags.Carry = value;
    value >>= 1;
    WriteRegisterW ( reg, value );
    CompareW ( value, 0 );
}

void opcode_SLA ()
{
    int reg = curOpCode & 0x000F;
    int count = ( curOpCode >> 4 ) & 0x000F;
    if ( count == 0 ) {
        count = ReadRegisterW ( 0 ) & 0x000F;
        if ( count == 0 ) count = 16;
    }
    USHORT value = ReadRegisterW ( reg );
    USHORT mask = 0xFFFF << ( 16 - count );
    USHORT overflow = value & mask;
    ST.flags.Overflow = (( overflow != mask ) && ( overflow != 0 )) ? 1 : 0;
    ST.flags.Carry = ( value & ( 0x8000 >> ( count - 1 ))) ? 1 : 0;
    value <<= count;
    WriteRegisterW ( reg, value );
    CompareW ( value, 0 );
}

void opcode_SRC ()
{
    int reg = curOpCode & 0x000F;
    short count = ( curOpCode >> 4 ) & 0x000F;
    if ( count == 0 ) {
        count = ReadRegisterW ( 0 ) & 0x000F;
        if ( count == 0 ) count = 16;
    }
    USHORT value = ReadRegisterW ( reg );
    ST.flags.Carry = ( value & ( 0x8000 >> ( count - 1 ))) ? 1 : 0;
    _CL = count;
    _AX = value;
    asm ror ax,cl;
    value = _AX;
    WriteRegisterW ( reg, value );
    CompareW ( value, 0 );
}

void opcode_CKOF () {}
void opcode_CKON () {}
void opcode_IDLE () {}
void opcode_LREX () {}
void opcode_RSET () {}

void CompareW ( USHORT arg1, USHORT arg2 )
{
    _AX = arg1;
    _DX = arg2;
l	cmp	ax,dx;
l	seta	al;			// Logic Greater
l	setg	ah;			// Arith Greater
l	sete	dl;			// Equal
l	shl	al,1;
l	or	al,ah;
l	shl	al,1;
l	or	al,dl;
l	shl	ax,13;
    ST.value &= 0x1FFF;
    ST.value |= _AX;
}

void CompareB ( UCHAR arg1, UCHAR arg2 )
{
    _AL = arg1;
    _DL = arg2;
l	cmp	al,dl;
l	seta	al;			// Logic Greater
l	setg	ah;			// Arith Greater
l	sete	dl;			// Equal
l	shl	al,1;
l	or	al,ah;
l	shl	al,1;
l	or	al,dl;
l	shl	al,3;
    _AH = arg1;
l	cmp	ah,0
l	setpo	ah			// Parity
l	or   	al,ah
l	shl	ax,10;
    ST.value &= 0x1BFF;
    ST.value |= _AX;
}

USHORT GetAddress ()
{
    USHORT retVal = 0;
    int reg = curOpCode & 0x0F;
    switch ( curOpCode & 0x0030 ) {
        case 0x0000 : retVal = WP + 2 * reg;			break;
        case 0x0010 : retVal = ReadRegisterW ( reg );		break;
        case 0x0030 : retVal = ReadRegisterW ( reg );
                      WriteRegisterW ( reg, ( USHORT ) ( retVal + size ));
                      break;
        case 0x0020 : if ( reg ) retVal = ReadRegisterW ( reg );
                      retVal += *PC++;
                      break;
    }
    return retVal;
}

USHORT Prefetch ()
{
    USHORT data = _ReadMemoryW_ ( PC );
    PC += 2;
    return data;
}

void CheckInterrupt ()
{
    int level = InterruptFlag;
    if ( level > ST.flags.IntMask ) return;
    prefetchOK = false;
    InterruptFlag = -1;
    USHORT newWP = _ReadMemoryW_ ( level * 4 );
    USHORT newPC = _ReadMemoryW_ ( level * 4 + 2 );
    ContextSwitch ( newWP, newPC );
    if ( level ) ST.flags.IntMask = level - 1;
}

void ContextSwitch ( ADDRESS newWP, ADDRESS newPC )
{
    USHORT oldWP = WP;
    USHORT oldPC = PC;
    WP = newWP;
    PC = newPC;
    WriteRegisterW ( 13, oldWP );
    WriteRegisterW ( 14, oldPC );
    WriteRegisterW ( 15, ST.value );
}

void ExecuteInstruction ()
{
    sLookUp *lookup = &LookUp [ curOpCode >> 12 ];
    op = lookup->opCode;
    int retries = lookup->size;
    if ( retries ) {
        while (( curOpCode & op->mask ) != op->opCode ) {
            op++;
            if ( retries-- == 0 ) return;
        }
    }
    size = 2;
    switch ( op->format ) {
        case 1 :				// Two General Addresses
                 if (( curOpCode & 0x0030 ) == 0x0020 ) arg1 = Prefetch ();
                 if (( curOpCode & 0x0C00 ) == 0x0800 ) arg2 = Prefetch ();
                 if ( curOpCode & 0x1000 ) size = 1;
                 break;
        case 3 :				// Logical
        case 4 :				// CRU Mult-Bit
        case 6 :				// Single Address
        case 9 :				// XOP, MULT, & DIV
                 if (( curOpCode & 0x0030 ) == 0x0020 ) arg1 = Prefetch ();
                 break;
        case 8 :				// Immediate
                 arg1 = Prefetch ();
    }
}

void PrefetchInstruction ()
{
    curOpCode = Prefetch ();
    ExecuteInstruction ();
    prefetchOK = true;
}

bool Step ()
{
    CheckInterrupt ();
    if ( ! prefetchOK ) PrefetchInstruction ();
    ((void(*)(void*)) op->function )( this );
    prefetchOK = false;
    InstructionCounter++;
    return false;
}
